/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// sv_edict.c -- entity dictionary

#include	"../quakedef.h"
#include	"../game.h"

#define		FIELDDEF_COUNT	10

ddef_t pr_fielddefs[FIELDDEF_COUNT] =
{
	{ev_vector,	FOFS(absmin),		"absmin"},
	{ev_vector,	FOFS(absmax),		"absmax"},
	{ev_vector,	FOFS(angles),		"angles"},
	{ev_vector,	FOFS(origin),		"origin"},

	{ev_float,	FOFS(number),		"number"},
	{ev_float,	FOFS(modelindex),	"modelindex"},

	{ev_string,	FOFS(wad),			"wad"},
	{ev_string,	FOFS(model),		"model"},
	{ev_string,	FOFS(message),		"message"},
	{ev_string,	FOFS(classname),	"classname"}
};

qboolean	ED_ParseEpair (void *base, ddef_t *key, char *s);

/*
=============
ED_ClearEdict
=============
*/
void ED_ClearEdict (edict_t *e)
{
	MR_memset (&e->v,		0, sizeof(e->v));
	MR_memset (&e->game,	0, sizeof(e->game));
	e->free = false;
}

/*
========
ED_Alloc
========
*/
edict_t *ED_Alloc (void)
{
	int			i;
	edict_t		*e;

	for (i=svs.maxclients+1; i<sv.num_edicts; i++)
	{
		e = EDICT_NUM(i);
		// the first couple seconds of server time can involve a lot of
		// freeing and allocating, so relax the replacement policy
		if (e->free && ( e->freetime < 2 || sv.time - e->freetime > 0.5 ) )
		{
			ED_ClearEdict (e);
			return e;
		}
	}
	
	if (i == MAX_EDICTS)
		Sys_Error ("ED_Alloc: no free edicts");
		
	sv.num_edicts++;
	e = EDICT_NUM(i);
	ED_ClearEdict (e);

	return e;
}

/*
=======
ED_Free
=======
*/
void ED_Free (edict_t *ed)
{
	SV_UnlinkEdict (ed);

	ed->free			= true;
	ed->v.model			= 0;
	ed->v.modelindex	= 0;
	VectorCopy (vec3_origin, ed->game.origin);
	VectorCopy (vec3_origin, ed->v.angles);
	ed->game.nextthink	= -1;
	ed->game.solid		= 0;
	
	ed->freetime = sv.time;
}

/*
============
ED_FindField
============
*/
ddef_t *ED_FindField (char *name)
{
	ddef_t		*def;
	int			i;
	
	for (i=0 ; i<ENTVAR_COUNT; i++)
	{
		def = &pr_fielddefs[i];
		if (!MR_strcmp (def->s_name,name) )
			return def;
	}
	return NULL;
}

/*
==============================================================================

					ARCHIVING GLOBALS

==============================================================================
*/

/*
=============
ED_NewString
=============
*/
char *ED_NewString (char *string)
{
	char	*new_string, *new_p;
	int		i,l;
	
	l = MR_strlen (string) + 1;
	new_string = Hunk_AllocName (l, "unknown");
	new_p = new_string;

	for (i=0 ; i< l ; i++)
	{
		if (string[i] == '\\' && i < l-1)
		{
			i++;
			if (string[i] == 'n')
				*new_p++ = '\n';
			else
				*new_p++ = '\\';
		}
		else
			*new_p++ = string[i];
	}
	
	return new_string;
}

/*
=============
ED_ParseEpair
=============
*/
qboolean ED_ParseEpair (void *base, ddef_t *key, char *s)
{
	int		i;
	char	string[128];
	ddef_t	*def;
	char	*v, *w;
	void	*d;
	
	d = (void *)((int *)base + key->ofs);
	
	switch (key->type)
	{
	case ev_string:
		*((char **)d) = ED_NewString(s);
		break;
		
	case ev_float:
		*(float *)d = MR_atof (s);
		break;
		
	case ev_vector:
		MR_strcpy (string, s);
		v = string;
		w = string;
		for (i=0 ; i<3 ; i++)
		{
			while (*v && *v != ' ')
				v++;
			*v = 0;
			((float *)d)[i] = MR_atof (w);
			w = v = v+1;
		}
		break;
		
	case ev_entity:
		*(int *)d = EDICT_TO_PROG(EDICT_NUM(MR_atoi (s)));
		break;
		
	case ev_field:
		def = ED_FindField (s);
		if (!def)
		{
			Con_Printf ("Can't find field %s\n", s);
			return false;
		}
		break;
	
	case ev_function:
		Con_Printf ("Can't find function %s\n", s);
		return false;
		break;
		
	default:
		break;
	}
	return true;
}

/*
=============
ED_ParseEdict
=============
*/
char *ED_ParseEdict (char *data, edict_t *ent)
{
	ddef_t		*key;
	qboolean	anglehack;
	qboolean	init;
	char		keyname[256];
	int			n;

	init = false;

	if (ent != sv.edicts)
		MR_memset (&ent->v, 0, sizeof(ent->v));

	while (1)
	{	
		data = COM_Parse (data);
		if (com_token[0] == '}')
			break;
		if (!data)
			Sys_Error ("ED_ParseEntity: EOF without closing brace");
		
		if (!MR_strcmp (com_token, "angle"))
		{
			MR_strcpy (com_token, "angles");
			anglehack = true;
		}
		else
			anglehack = false;

		MR_strcpy (keyname, com_token);

		n = MR_strlen (keyname);
		while (n && keyname[n-1] == ' ')
		{
			keyname[n-1] = 0;
			n--;
		}

		data = COM_Parse (data);
		if (!data)
			Sys_Error ("ED_ParseEntity: EOF without closing brace");

		if (com_token[0] == '}')
			Sys_Error ("ED_ParseEntity: closing brace without data");

		init = true;	

		if (keyname[0] == '_')
			continue;
		
		key = ED_FindField (keyname);
		if (!key)
		{
			Con_Printf ("'%s' is not a field\n", keyname);
			continue;
		}

		if (anglehack)
		{
			char	temp[32];
			MR_strcpy (temp, com_token);
			_snprintf (com_token, sizeof(com_token), "0 %s 0", temp);
		}

		if (!ED_ParseEpair ((void *)&ent->v, key, com_token))
			Host_Error ("ED_ParseEdict: parse error");
	}

	if (!init)
		ent->free = true;

	return data;
}

/*
===============
ED_LoadFromFile
===============
*/
void ED_LoadFromFile (char *data)
{	
	edict_t		*ent;
	int			i;
	qboolean	foundit;
	
	ent = NULL;
	
	while (1)
	{
		data = COM_Parse (data);
		if (!data)
			break;
		if (com_token[0] != '{')
			Sys_Error ("ED_LoadFromFile: found %s when expecting {",com_token);

		if (!ent)
			ent = EDICT_NUM(0);
		else
			ent = ED_Alloc ();

		data = ED_ParseEdict (data, ent);

		if (!ent->v.classname)
		{
			Con_Printf ("Something failed\n");
			ED_Free (ent);
			continue;
		}

		ent->game.classname = ent->v.classname;

		foundit = false;

		for (i=0; i<NUM_ENGINE_FUNCS; i++)
		{
			if (!MR_strcmp (gamefuncs[i].name, ent->game.classname))
			{
				miniracer.self = ent;
				gamefuncs[i].func ();
				foundit = true;
				continue;
			}
		}

		if (!foundit)
		{
			Con_Printf ("Something failed\n");
			ED_Free (ent);
		}
	}
}

edict_t *EDICT_NUM(int n)
{
	if (n < 0 || n >= sv.max_edicts)
		Sys_Error ("EDICT_NUM: bad number %i", n);
	return (edict_t *)((byte *)sv.edicts+ (n)*sizeof(edict_t));
}

int NUM_FOR_EDICT(edict_t *e)
{
	int		b;
	
	b = (byte *)e - (byte *)sv.edicts;
	b = b / sizeof(edict_t);
	
	if (b < 0 || b >= sv.num_edicts)
		Sys_Error ("NUM_FOR_EDICT: bad pointer");
	return b;
}
